import logging
import sys
import time
import threading

from queue import Queue
from .methods import *  # NOQA


class ledPulse(threading.Thread):

    delayMethods = ['constant', 'sin', 'cos']
    valueMethods = ['on', 'off', 'linear', 'vshape', 'sin', 'cos']
    methods = {'delay': {}, 'value': {}}

    defaultSet = {
        'min': 2,
        'max': 50,
        'delay': 0.01,
        'delayMethod': 'constant',
        'initialMethod': None,
        'loopMethod': 'linear',
        'finalMethod': None
    }

    def __init__(self, gpio, name="led0"):

        super().__init__(name=name,
                         daemon=True)

        self.log = logging.getLogger(self.__class__.__name__)
        self.gpio = gpio

        self.pwm_setup()

        self.log.info(f"platform '{self.model}' " +
                      f"support: {self.supported}")

        for dm in self.delayMethods:
            self.register_delay_method(dm, eval(f"delay_{dm}"))

        for vm in self.valueMethods:
            self.register_value_method(vm, eval(f"value_{vm}"))

        self.stop_event = threading.Event()

        # Method to stop thread
        # Like
        # def stop():
        #    self.stop_event.set()
        # But shorter ;)
        self.stop = self.stop_event.set

        self.queue = Queue()

        # Set default values
        for key, value in self.defaultSet.items():
            setattr(self, key, value)

    def pwm_setup(self):
        """
        We check if we are running on RPi.
        If yes we set supported to True and set setValue method
        to changeDutyCycle, else
        we set supported to False and create a dummy setValue method
        """

        self.supported = False
        self.model = sys.platform
        try:
            with open('/sys/firmware/devicetree/base/model', 'r') as m:
                model = m.read().strip()
                if model.lower().startswith('raspberry pi'):
                    self.supported = True
                    self.model = model
        except Exception:
            pass

        if self.supported:
            import RPi.GPIO as GPIO
            GPIO.setwarnings(False)          # disable warnings
            GPIO.setmode(GPIO.BCM)           # set pin numbering system
            GPIO.setup(self.gpio, GPIO.OUT)  # set GPIO for output

            # create PWM instance with 120Hz frequency
            self.pwm = GPIO.PWM(self.gpio, 120)
            self.pwm.start(0)  # start pwm with value 0, off
            # setValue function
            self.setValue = self.pwm.ChangeDutyCycle
        else:
            # Platform not supported, set
            # a dummy setValue function
            if self.log.root.level == logging.DEBUG:
                self.setValue = print
            else:
                self.setValue = lambda x: x

        return (self.supported, sys.platform)

    def set(self, **kwargs):

        for key, value in kwargs.items():
            if key in self.defaultSet.keys():
                setattr(self, key, value)
            else:
                self.log.warning(f"set: unknown parameter '{key}'")

        self.log.debug("calculating values: "
                       f"initialMethod:{self.initialMethod}, "
                       f"loopMethod:{self.loopMethod}, "
                       f"finalMethod:{self.finalMethod}, "
                       f"delayMethod:{self.delayMethod}, min:{self.min}, "
                       f"max:{self.max}, delay:{self.delay}")

        delayValues = []
        for step in range(0, 100):
            try:
                delay = self.methods['delay'][self.delayMethod](self, step)
            except NameError:
                self.log.error(f"delay method '{self.delayMethod}' not found")
                return
            delayValues.append(delay)

        initialValues = []
        if self.initialMethod:
            for step in range(0, 50):
                try:
                    value = self.methods['value'][self.initialMethod](self,
                                                                      step)
                except KeyError:
                    self.log.error(f"value method '{self.initialMethod}' "
                                   "not found")
                    return
                initialValues.append((value, delayValues[step]))

        loopValues = []
        for step in range(0, 100):
            try:
                value = self.methods['value'][self.loopMethod](self, step)
            except KeyError:
                self.log.error(f"value method '{self.loopMethod}' not found")
                return
            loopValues.append((value, delayValues[step]))

        finalValues = []
        if self.finalMethod:
            for step in range(50, 100):
                try:
                    value = self.methods['value'][self.finalMethod](self, step)
                except KeyError:
                    self.log.error(f"value method '{self.finalMethod}' "
                                   "not found")
                    return
                finalValues.append((value, delayValues[step]))

        self.queue.put((initialValues, loopValues, finalValues))

    def run(self):

        while True:
            if self.stop_event.is_set():
                self.log.debug("break from main loop, bye")
                if self.supported:
                    self.pwm.stop()
                break

            # Get an item from the queue,
            # if queue is empty, wait for a new item
            # item is a tuple with (initialValues, loopValues, finalValues)
            (initialValues, loopValues, finalValues) = self.queue.get()

            # Initial loop
            self.log.debug('running initial steps')
            for value, delay in initialValues:
                self.setValue(value)
                time.sleep(delay)

            self.log.debug('running main loop')
            while True:
                for value, delay in loopValues:
                    self.setValue(value)
                    time.sleep(delay)

                # After each loop we check if there is something waiting in
                # queue or if stop has been raised
                if not self.queue.empty() or self.stop_event.is_set():
                    break

            self.log.debug('running final steps')
            for value, delay in finalValues:
                self.setValue(value)
                time.sleep(delay)

    ###
    # PLUGINS METHODS
    # We register new functions for delay or value
    # Function must get `step` variable, specifing at which step
    # in the loop we are and is added as a method to self, so
    # it has access to all the properties of self: max, min, delay
    ###
    def register_value_method(self, name, fn):
        self.register_method('value', name, fn)

    def register_delay_method(self, name, fn):
        self.register_method('delay', name, fn)

    def register_method(self, kind, name, fn):
        if kind not in ['delay', 'value']:
            self.log.warning(f"unknown kind {kind}")
            return

        if name not in self.methods[kind]:
            self.methods[kind][name] = fn
            self.log.info(f"registered {kind} method '{name}'")
        else:
            self.log.warning(f"{kind} method '{name}' already defined")



import math

"""
Predefined methods
"""

"""
Delay methods
"""


def delay_sin(obj, step):
    return math.sin(math.radians(1.8*step)) * obj.delay


def delay_cos(obj, step):
    return abs(math.cos(math.radians(1.8*step))) * obj.delay


def delay_constant(obj, step):
    return obj.delay


"""
Values methods
"""


def value_on(obj, step):
    """
    Always on at "max" brightness
    """

    return obj.max


def value_off(obj, step):
    """
    Always off
    """

    return 0


def value_linear(obj, step):
    """
    Saw patterned value, 0-1-0
    """

    delta = obj.max - obj.min
    if step < 50:
        return delta*step/50 + obj.min
    else:
        return delta * (100-step)/50 + obj.min


def value_vshape(obj, step):
    """
    V-shape value, 1-0-1
    """

    delta = obj.max - obj.min
    if step < 50:
        return delta * (100-step)/50 + obj.min
    else:
        return delta*step/50 + obj.min


def value_sin(obj, step):
    """
    Sinusoidal values, 0-1-0
    """

    delta = obj.max - obj.min
    radians = math.radians(1.8 * step)
    return delta * math.sin(radians) + obj.min


def value_cos(obj, step):
    """
    Absolute Cosinusoidal values, 1-0-1
    """

    delta = obj.max - obj.min
    radians = math.radians(1.8 * step)
    return delta * abs(math.cos(radians)) + obj.min



"""
Usage:
    pulses [--gpio <gpio>] [--initial <initialMethod>] [--loop <loopMethod>]
                           [--final <finalMethod>] [--delay <delayMethod>]
                           [--min <min>] [--max <max>] [--delay-val <delay>]
                           [--verbose]

Options:
    -g --gpio=<gpio>         GPIO to use [default: 12]
    -i --initial=<method>    Initial method [default: linear]
    -l --loop=<method>       Loop method [default: sin]
    -f --final=<method>      Final method [default: linear]
    -d --delay=<method>      Delay method [default: constant]

    -m --min=<min>           Minimum value [default: 2]
    -M --max=<max>           Maximum value [default: 50]
    -D --delay-val=<delay>   Base delay value [default: 0.01]

    -V --verbose             Verbose mode [default: True]

    -h --help                Show help.
    -v --version             Show version.

""" # NOQA

import sys
import time
import logging
import signal
from docopt import docopt
from pulses import VERSION, ledPulse

logging.basicConfig(format="%(asctime)-15s %(levelname)5s %(name)s[%(process)d] %(threadName)s: %(message)s")


def signal_handler(sig, frame):
    print('stopping pulse...', end="", flush=True)
    led.stop()
    led.join()
    print('done.\nBailing out')
    sys.exit(0)


def main():
    global led

    args = docopt(__doc__, version=f'pulses {VERSION}')
    signal.signal(signal.SIGINT, signal_handler)

    log = logging.getLogger()
    log.setLevel('WARNING')
    if args.get('--verbose'):
        log.setLevel('DEBUG')
        log.info('setting verbose mode')

    led = ledPulse(int(args.get('--gpio')))

    params = {}
    for method in ['initial', 'loop', 'final']:
        methodName = args.get(f"--{method}")
        if methodName not in led.valueMethods:
            print(f"error: no value method '{methodName}' defined")
            sys.exit(1)
        params[f"{method}Method"] = methodName

    methodName = args.get("--delay")
    if methodName not in led.delayMethods:
        print(f"error: no delay method '{methodName}' defined")
        sys.exit(1)
    params['delayMethod'] = methodName

    params['min'] = int(args.get('--min'))
    params['max'] = int(args.get('--max'))
    params['delay'] = float(args.get('--delay-val'))

    print("-" * 20)
    print("pulses CLI test tool")
    print("-" * 20)
    print("\n> loop parameters")
    for param, value in params.items():
        print(f"  {param}: {value}")
    led.set(**params)
    print('\nstarting loop, press Ctrl-C to stop...')
    led.start()
    while True:
        time.sleep(4)


if __name__ == "__main__":
    main()
